@@ -57,12 +57,13 @@ PUBLIC void HTFormatInit NOARGS
  if ((cp = getenv(DISPLAY)) != NULL && *cp != '\0') {	/* Must have X11 */
   HTSetPresentation("application/postscript", "ghostview %s&",
   							    1.0, 3.0, 0.0, 0);
-  HTSetPresentation("image/gif",        XLOADIMAGE_COMMAND, 1.0, 3.0, 0.0, 0);
-  HTSetPresentation("image/x-xbm",      XLOADIMAGE_COMMAND, 1.0, 3.0, 0.0, 0);
-  HTSetPresentation("image/x-png",      XLOADIMAGE_COMMAND, 1.0, 3.0, 0.0, 0);
-  HTSetPresentation("image/x-rgb",      XLOADIMAGE_COMMAND, 1.0, 3.0, 0.0, 0);
-  HTSetPresentation("image/x-tiff",     XLOADIMAGE_COMMAND, 1.0, 3.0, 0.0, 0);
-  HTSetPresentation("image/jpeg",       XLOADIMAGE_COMMAND, 1.0, 3.0, 0.0, 0);
+  HTSetPresentation("image/gif",        XLoadImageCommand,  1.0, 3.0, 0.0, 0);
+  HTSetPresentation("image/x-xbm",      XLoadImageCommand,  1.0, 3.0, 0.0, 0);
+  HTSetPresentation("image/x-xbitmap",  XLoadImageCommand,  1.0, 3.0, 0.0, 0);
+  HTSetPresentation("image/x-png",      XLoadImageCommand,  1.0, 3.0, 0.0, 0);
+  HTSetPresentation("image/x-rgb",      XLoadImageCommand,  1.0, 3.0, 0.0, 0);
+  HTSetPresentation("image/x-tiff",     XLoadImageCommand,  1.0, 3.0, 0.0, 0);
+  HTSetPresentation("image/jpeg",       XLoadImageCommand,  1.0, 3.0, 0.0, 0);
   HTSetPresentation("video/mpeg",       "mpeg_play %s &",   1.0, 3.0, 0.0, 0);
 
  }
@@ -141,6 +142,12 @@ PUBLIC void HTFormatInit NOARGS
 /*
  *  Now add our basic conversions.
  */
+ HTSetConversion("text/x-sgml",
+ 			      "www/source",  HTPlainPresent, 1.0, 0.0, 0.0, 0);
+ HTSetConversion("text/x-sgml",
+ 			      "www/present", HTMLPresent,    1.0, 0.0, 0.0, 0);
+ HTSetConversion("text/sgml", "www/source",  HTPlainPresent, 1.0, 0.0, 0.0, 0);
+ HTSetConversion("text/sgml", "www/present", HTMLPresent,    1.0, 0.0, 0.0, 0);
  HTSetConversion("text/plain","www/present", HTPlainPresent, 1.0, 0.0, 0.0, 0);
  HTSetConversion("text/html", "www/source",  HTPlainPresent, 1.0, 0.0, 0.0, 0);
  HTSetConversion("text/html", "text/x-c",    HTMLToC, 	     0.5, 0.0, 0.0, 0);
@@ -391,7 +398,7 @@ PRIVATE int ProcessMailcapEntry ARGS2(FILE *,fp, struct MailcapEntry *,mc)
 		/* ExceptionalNewline(mc->contenttype, atoi(eq)); */
 	    } else if (eq && !strcmp(arg, "q")) {
 	        mc->quality = atof(eq);
-		if (mc->quality < 0.001)
+		if (mc->quality > 0.000 && mc->quality < 0.001)
 		    mc->quality = 0.001;
 	    } else if (eq && !strcmp(arg, "mxb")) {
 	        mc->maxbytes = atol(eq);
@@ -484,6 +491,7 @@ PRIVATE int PassesTest ARGS1(struct MailcapEntry *,mc)
      *  Save overhead of system() calls by faking these. - FM
      */
     if (0 == strcasecomp(mc->testcommand, "test -n \"$DISPLAY\"")) {
+        FREE(mc->testcommand);
         if (TRACE)
 	    fprintf(stderr,"Testing for XWINDOWS environment.\n");
     	if ((cp = getenv(DISPLAY)) != NULL && *cp != '\0') {
@@ -497,6 +505,7 @@ PRIVATE int PassesTest ARGS1(struct MailcapEntry *,mc)
 	}
     }
     if (0 == strcasecomp(mc->testcommand, "test -z \"$DISPLAY\"")) {
+        FREE(mc->testcommand);
         if (TRACE)
 	    fprintf(stderr,"Testing for NON_XWINDOWS environment.\n");
     	if (!((cp = getenv(DISPLAY)) != NULL && *cp != '\0')) {
@@ -514,6 +523,7 @@ PRIVATE int PassesTest ARGS1(struct MailcapEntry *,mc)
      *  Why do anything but return success for this one! - FM
      */
     if (0 == strcasecomp(mc->testcommand, "test -n \"$LYNX_VERSION\"")){
+        FREE(mc->testcommand);
         if (TRACE) {
 	    fprintf(stderr,"Testing for LYNX environment.\n");
 	    fprintf(stderr,"[HTInit] Test passed!\n");
@@ -524,6 +534,7 @@ PRIVATE int PassesTest ARGS1(struct MailcapEntry *,mc)
      *  ... or failure for this one! - FM
      */
     if (0 == strcasecomp(mc->testcommand, "test -z \"$LYNX_VERSION\"")) {
+        FREE(mc->testcommand);
         if (TRACE) {
 	    fprintf(stderr,"Testing for non-LYNX environment.\n");
 	    fprintf(stderr,"[HTInit] Test failed!\n");
